#!/usr/bin/env node

const fs = require('fs');
const path = require('path');
const request = require('request');
const url = require('url');
const extract = require('extract-zip');
const tmp = require('tmp');
const del = require('del');
const maxErrorLen = 2048;

tmp.setGracefulCleanup();
const packageInfo = require(path.resolve('package.json'));
const nodeModulesPath = path.resolve('./node_modules/');

const envFilePath = path.resolve('.env');
if (fs.existsSync(envFilePath)) {
  console.debug('Using environment config file:', envFilePath);
  require('dotenv').config({path: envFilePath});
}

function installingAsDependency() {
  // ./main-package/node_modules/dependency/package.json
  // ./main-package/package.json
  const rootPackagePath = path.resolve('..', '..', 'package.json');
  if (fs.existsSync(rootPackagePath)) {
    const rootPackageInfo = require(rootPackagePath);
    return rootPackageInfo.name !== packageInfo.name;
  } else {
    return false;
  }
}

if (!packageInfo.jazDependencies) {
  console.debug('There aren\'t JAZ dependencies in package.json.');
} else if (installingAsDependency()) {
  console.debug('JAZ Packages were not installed because this package was installed ' +
  'as dependency');
} else {
  console.info(`Installing JAZ dependencies to package ${packageInfo.name}`);
  if (!process.env.SERVER_URL || !process.env.JAZ_API_KEY) {
    const name = process.env.SERVER_URL ? 'JAZ_API_KEY' : 'SERVER_URL';
    throw new Error('The environment variable "' + name + '" was not informed. Inform this ' +
      'environment variable or configure an ".env" config file in the project root.');
  }

  packageInfo.jazDependencies.forEach(packageName => {
    const packageDir = path.resolve(nodeModulesPath, packageName);
    if (!fs.existsSync(packageDir)) {
      fs.mkdirSync(packageDir);
      console.info('Created directory', packageDir);
    }
    const eTagFileName = path.resolve(packageDir, '.etag');
    let currentETag = '';
    if (fs.existsSync(eTagFileName)) {
      currentETag = fs.readFileSync(eTagFileName, 'utf8');
    }

    const tempFileName = tmp.tmpNameSync();
    let statusCode = 0;
    let requestETag = '';
    request
      .get(url.resolve(process.env.SERVER_URL, '/api/jaz/v1/packages/' + packageName), {
        qs: {
          api_key: process.env.JAZ_API_KEY
        },
        headers: {
          'if-none-match': currentETag
        }
      })
      .on('response', function(response) {
        statusCode = response.statusCode;
        requestETag = response.headers.etag || '';
      })
      .pipe(fs.createWriteStream(tempFileName))
      .on('finish', () => {
        if (statusCode === 200) {
          del.sync(packageDir + '/**', '!' + packageDir, '!' + eTagFileName);
          extract(tempFileName, {dir: packageDir}, function (err) {
            if (err) {
              console.error(`Error extracting JAZ file ${packageName} content:`, err);
              process.exitCode = 1;
            } else {
              console.info(`JAZ file ${packageName} extracted successfully`);
              fs.writeFileSync(eTagFileName, requestETag);
            }
          });
        } else if (statusCode === 304) {
          console.info(`Jaz dependency ${packageName} is up to date`);
        } else if (statusCode === 404) {
          console.error(`Jaz dependency ${packageName} not found`);
        } else {
          const body = fs.readFileSync(tempFileName).toString();
          let msg;
          try {
            const error = JSON.parse(body);
            msg = error.message;
          } catch (e) {
            msg = body.substr(0, maxErrorLen);
          }
          console.error(`Error downloading dependency ${packageName} (${statusCode}): ${msg}`);
          process.exitCode = 1;
        }
      });
  });
}
