#!/usr/bin/env node

const fs = require('fs');
const path = require('path');
const glob = require('glob');
const archiver = require('archiver');
const UglifyJS = require('uglify-es');
const babel = require("babel-core");
const prettyBytes = require('pretty-bytes');
const mkdirp = require('mkdirp');
const zlib = require('zlib');
const E2cFile = require('../lib/E2cFile');

const e2cPublicKey =
  '-----BEGIN PUBLIC KEY-----\n' +
  'MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAw7nmFoo1k1pYRe0w9lQC\n' +
  '5yXL6pUu3kNCIhRAz5krYFFf/JFATM22NRaoR4sN9+r9X3lymb/B7QEJXDv2V+sP\n' +
  'B+9eSWUIUU8FhXuOobyE2XSpaC7DJfpNPC3oFTTHRVTJ/zUIJyYEztHAddNlhWo6\n' +
  'c/PTq928wBhXXLTkO+v/EIk9egKcfDCPbqLAy3k5ssU4R6GDhv7RgYnK/LvouKDr\n' +
  'ehN9oiYZVJDylaPzucVHKL3eqJv+BRB8EE9vAuC6/BpsIKbE7qaITMr/2OSio6qe\n' +
  'CvuijGi1ztfhedeEmC0cChlhy0rWSoILvcn3Ge8p+y6szNu5iquOvD9Cnp5qdkwn\n' +
  'xwIDAQAB\n' +
'-----END PUBLIC KEY-----';

const babelOptions = {
  sourceType: "script",
  presets: [],
  plugins: [
    ["transform-es2015-block-scoping", {"throwIfClosureRequired": true}]
  ]
};

const envFilePath = path.resolve('.env');
if (fs.existsSync(envFilePath)) {
  console.debug('Using environment config file:', envFilePath);
  require('dotenv').config({path: envFilePath});
}

const configPath = process.argv[2];
if (!configPath) {
  throw new Error('JAZ package config not informed.\n' +
    'Usage: packJaz jaz.config.json');
}

const config = require(path.resolve(configPath));
const basePath = path.dirname(configPath);

const filesToArchive = new Set();
const filesToUglify = new Set();
const filesToEncrypt = new Set();
const filesMetadata = new Map();

if (config.files) {
  config.files
    .map(pattern => glob.sync(pattern, {cwd: basePath}))
    .forEach(matches => matches.forEach(name => filesToArchive.add(name)));
}

if (config.ignore) {
  config.ignore
    .map(pattern => glob.sync(pattern, {cwd: basePath}))
    .forEach(matches => matches.forEach(name => filesToArchive.delete(name)));
}

if (config.uglify && config.uglify.files) {
  config.uglify.files
    .map(pattern => glob.sync(pattern, {cwd: basePath}))
    .forEach(matches => matches.forEach(name => filesToUglify.add(name)));
}

if (config.e2c) {
  const rules = config.e2c instanceof Array ? config.e2c : [config.e2c];
  rules.forEach(rule => {
    const metadata = Buffer.from(JSON.stringify(rule.metadata || {}, 'utf-8'));
    rule.files
      .map(pattern => glob.sync(pattern, {cwd: basePath}))
      .forEach(matches => matches.forEach(name => {
        filesToEncrypt.add(name);
        filesMetadata.set(name, metadata);
      }));
  });
}

const uglifyOptions = (config.uglify && config.uglify.options) || {};

let destPath = config.dest;
if (!path.isAbsolute(destPath)) {
  destPath = path.resolve(basePath, destPath);
}

mkdirp.sync(path.dirname(destPath));

const destStream = fs.createWriteStream(destPath);
const archive = archiver('zip', {
  zlib: {
    level: 9
  }
});

archive.on('warning', function (err) {
  if (err.code === 'ENOENT') {
    console.warn(err);
  } else {
    throw err;
  }
});

archive.on('error', function (err) {
  throw err;
});

let entries = 0;
archive.on('entry', function () {
  entries++;
});

destStream.on('close', function () {
  console.log(`${config.dest} built with ${entries} entries and ${prettyBytes(archive.pointer())}`);
});

archive.pipe(destStream);

archive.append('{"encoding": "UTF-8"}', {
  name: 'jazmanifest.json',
  date: new Date(2017, 3, 30)
});

for (const relativePath of filesToArchive) {
  const filePath = path.resolve(basePath, relativePath);
  const fstat = fs.statSync(filePath);

  const entryData = {
    name: process.platform === 'win32' ? relativePath.replace(/\\/g, '/') : relativePath,
    stats: fstat
  };

  if (fstat.isFile()) {
    let source = null;
    if (filesToUglify.has(relativePath)) {
      const out = UglifyJS.minify(
        babel.transform(fs.readFileSync(filePath, 'utf8'), babelOptions).code,
        uglifyOptions);
      if (!out.error) {
        source = out.code;
      } else {
        throw out.error;
      }
    }

    if (filesToEncrypt.has(relativePath)) {
      let contentBuffer;
      if (source === null) {
        contentBuffer = zlib.gzipSync(fs.readFileSync(filePath));
      } else {
        contentBuffer = zlib.gzipSync(Buffer.from(source, 'utf-8'));
      }
      source = E2cFile.encryptFile(
        e2cPublicKey,
        filesMetadata.get(relativePath),
        contentBuffer
      );
      entryData.name += '.gz.e2c';
    }

    if (source === null) {
      source = fs.createReadStream(filePath);
    }
    archive.append(source, entryData);
  }
}

archive.finalize();
