const fs = require('fs');
const path = require('path');
const assert = require('assert');
const request = require('request');
const url = require('url');
const EngineWrapper = require('./EngineWrapper');

const envFilePath = path.resolve('.env');
if (fs.existsSync(envFilePath)) {
  console.debug('Using environment config file:', envFilePath);
  require('dotenv').config({path: envFilePath});
}

const defaultJazArchivesPath = path.resolve('build/jaz');
const defaultEngineWorkDir = path.resolve('.engine');

class Builder {
  constructor () {
    assert.ok(process.env.SERVER_DB_NAME, 'Environment variable SERVER_DB_NAME is not defined');
    assert.ok(process.env.SERVER_URL, 'Environment variable SERVER_URL is not defined');
    assert.ok(process.env.SERVER_USR, 'Environment variable SERVER_USR is not defined');
    assert.ok(process.env.SERVER_PSW, 'Environment variable SERVER_PSW is not defined');
    assert.ok(process.env.BUILDER_API_KEY, 'Environment variable BUILDER_API_KEY is not defined');
    this.serverUrl = process.env.SERVER_URL;
    this.dbName = process.env.SERVER_DB_NAME;
    this.userName = process.env.SERVER_USR;
    this.userPassword = process.env.SERVER_PSW;
    this.builderApiKey = process.env.BUILDER_API_KEY;
    this.jazArchivesPath = defaultJazArchivesPath;
    this.engineWorkDir = defaultEngineWorkDir;
  }

  /**
   * Constructs a new Engine Wrapper configured to run build tasks.
   * @return {EngineWrapper}
   */
  newEngine() {
    const jazArchives = fs.existsSync(this.jazArchivesPath) ?
      fs.readdirSync(this.jazArchivesPath).map(name => path.resolve(this.jazArchivesPath, name)) :
      [];

    return new EngineWrapper(
      path.resolve(this.engineWorkDir, this.dbName),
      this.serverUrl,
      this.dbName,
      {jazArchives, disableIdoSyncToDisk: true}
    );
  }

  /**
   * Upload build artifacts to Engine Virtual File System. The allowed build artifacts must be
   * registered in /products/Development/configuration/System builder/.
   * @param {Array<{key: number, path: string}>} artifacts Artifacts to be uploaded.
   * @return {Promise}
   */
  uploadArtifacts(artifacts) {
    const engine = this.newEngine();
    return engine.start().then(() => {
      return Promise.all(artifacts.map((artifact) => {
        let artifactPath = artifact.path;
        const artifactStats = fs.statSync(artifactPath);
        if (!path.isAbsolute(artifactPath)) {
          artifactPath = path.resolve(artifactPath);
        }

        return new Promise((resolve, reject) => {
          console.info(`Uploading ${artifactPath} fo VFS file ${artifact.key}`);
          fs.createReadStream(artifactPath).pipe(
              request.post(url.resolve(engine.url, `/api/systemBuilder/v1/artifacts/${artifact.key}`), {
                qs: {
                  api_key: this.builderApiKey
                },
                headers: {
                  'content-length': artifactStats.size
                },
                auth: {
                  user: this.userName,
                  pass: this.userPassword,
                  sendImmediately: true
                }
              }, (err, response, body) => {
                if (err) {
                  reject(err);
                } else if (response.statusCode === 200) {
                  resolve();
                } else {
                  reject(new Error(`Invalid response from server. Status code: ${response.statusCode}: ` +
                      body));
                }
              })
          );
        });
      }));
    }).then(() => {
      const fileNames = artifacts.map(artifact => artifact.path).join(', ');
      console.info(`Files ${fileNames} uploaded successfully`);
      return engine.stop();
    }).catch(err => {
      const exitWithError = () => {
        console.error(err);
        return Promise.reject(err);
      };
      if (engine.running) {
        return engine.stop().then(exitWithError).catch(exitWithError);
      } else {
        return exitWithError();
      }
    });
  }

  /**
   * Run Engine server tests.
   * @param {string|Array<string>} testsPath Directory with the tests to be executed.
   * @param {string} reportPath Path to write the tests results report.
   * @return {Promise}
   */
  runServerTests(testsPath, reportPath) {
    const paths = (testsPath instanceof Array ? testsPath : [testsPath]).map(name => {
      return path.resolve(name);
    }).join(',');
    const engine = this.newEngine();
    let allTestsExecutedSuccessfully = false;
    return engine.start().then(() => {
      return new Promise((resolve, reject) => {
        console.info(`Running server-side tests in "${paths}"`);
        request.post(url.resolve(engine.url, '/api/systemBuilder/v1/tests/runLocal'), {
          qs: {
            api_key: this.builderApiKey,
            path: paths
          },
          auth: {
            user: this.userName,
            pass: this.userPassword,
            sendImmediately: true
          }
        }, (err, response, body) => {
          if (err) {
            console.error('Error running tests', err);
            reject(err);
          } if (response.statusCode === 200) {
            allTestsExecutedSuccessfully =
              response.headers['x-all-tests-executed-successfully'] === '1';
            console.info('Tests executed ' +
              (allTestsExecutedSuccessfully ? 'successfully' : 'with fails'));
            resolve(body);
          } else {
            reject(new Error(`Invalid response from server. Status code: ${response.statusCode}: ` +
              body));
          }
        });
      });
    }).then((body) => {
      return Promise.all([engine.stop(), new Promise((resolve, reject) => {
        fs.writeFile(reportPath, body, (err) => {
          if (err) {
            reject(err);
          } else {
            console.info(`Test results saved at ${reportPath}`);
            if (allTestsExecutedSuccessfully) {
              resolve();
            } else {
              reject(new Error('Tests failed'));
            }
          }
        });
      })]);
    }).catch(err => {
      const exitWithError = () => {
        console.error(err);
        return Promise.reject(err);
      };
      if (engine.running) {
        return engine.stop().then(exitWithError).catch(exitWithError);
      } else {
        return exitWithError();
      }
    });
  }
}

module.exports = Builder;
