const crypto = require('crypto');
const _ = require('struct-fu');

/* e2c file format

-- file header - plain text
[4] file format Id
[2] file version (little endian)
[2] metadata Size (little endian)
[n] metadata (size <= 64 Kb)
-- file header ext - plain binary
[32] header hash (sha256)
[16] authentication tag
[12] AES iv
[2] file prologue size + padding size (little endian)
-- end file header

----- start file prologue - RSA encrypted data using Engine public key
[4] payload size(little endian)
[16] AES 128 key
[n] padding
------ end file prologue

------ start file payload - AES GCM
[4] content size - (little endian)
[n] content
[] padding
-------end file payload
*/

const E2cFormatId = '#e2c';
const E2cVersion = 1;
const IVSize = 12;
const AES128KeySize = 16;

const FileHeaderStruct = _.struct([
  _.char('formatId', 4),
  _.uint16le('version'),
  _.uint16le('metadataSize')
	// uint8 metadata[n];
]);

const FileHeaderExtStruct = _.struct([
  _.uint8('headerHash', 32),
  _.uint8('authTag', 16),
  _.uint8('iv', IVSize),
  _.uint16le('prologueSize')
	// uint8 metadata[n];
]);

const FilePrologueStruct = _.struct([
	_.uint32le('payloadSize'),
	_.uint8('key', AES128KeySize)
	// uint8 padding[n]
]);

const FilePayloadStruct = _.struct([
	_.uint32le('contentSize')
	// uint8 content[n]
	// uint8 padding[n]
]);

function calcHeaderHash(header, metadata) {
  const hash = crypto.createHash('sha256');
  hash.update(FileHeaderStruct.pack(header));
  hash.update(metadata);
	return hash.digest();
}

module.exports = {

  /**
   * Encrypt a file content using Engine E2C file format.
   * @param {string} publicKey Engine RSA public key.
   * @param {Buffer} metadata File metadata. Used to informe license requirements to decrypt this
   * file. See Engine docs for more details.
   * @param {Buffer} content Original file content.
   * @return {Buffer} File encrypted content.
   */
  encryptFile: function (publicKey, metadata, content) {

    const header = {
      formatId: E2cFormatId,
      version: E2cVersion,
      metadataSize: metadata.length
    };

    const headerExt = {
      headerHash: calcHeaderHash(header, metadata),
      authTag: null,
      iv: new Uint8Array(IVSize),
      prologueSize: 0
    };
    crypto.randomFillSync(headerExt.iv);

    const prologue = {
      payloadSize: 0,
      key: new Uint8Array(AES128KeySize)
    };
    crypto.randomFillSync(prologue.key);

    const filePayload = {
      contentSize: content.length
    };

    const cipher = crypto.createCipheriv('aes-128-gcm', prologue.key, headerExt.iv);
    cipher.setAAD(headerExt.headerHash);
    const payload = Buffer.concat([
      cipher.update(FilePayloadStruct.pack(filePayload)),
      cipher.update(content),
      cipher.final()
    ]);

    prologue.payloadSize = payload.length;
    headerExt.authTag = cipher.getAuthTag();

    const encryptedPrologue = crypto.publicEncrypt({
      key: publicKey,
      padding: crypto.constants.RSA_PKCS1_PADDING
    }, FilePrologueStruct.pack(prologue));

    headerExt.prologueSize = encryptedPrologue.length;

    return Buffer.concat([
      FileHeaderStruct.pack(header),
      metadata,
      FileHeaderExtStruct.pack(headerExt),
      encryptedPrologue,
      payload
    ]);
  }
};
