const spawn = require('child_process').spawn;
const path = require('path');
const fs = require('fs');
const url = require('url');
const del = require('del');
const mkdirp = require('mkdirp');
const request = require('request');
const getPort = require('get-port');
const sleep = require('thread-sleep');

const ENGINE_APP = 'Engine64.exe';
const EXIT_CODE_RESTART_REQUIRED = 4;
const ENGINE_STATUS_URL = '/api/service-status/v1/status';
const ENGINE_DOWNLOAD_URL = `/dist/${ENGINE_APP}`;
const ENGINE_STARTUP_TIMEOUT = 15 * 60 * 1000;
const STATUS_MONITOR_INTERVAL = 500;
const UPDATE_INTERVAL = 30000;

class EngineWrapper {

  /**
   * Creates a new Engine wrapper that can be used to control a Bematech ERP Engine process.
   * @param {string} installPath Directory where the Engine will be installed.
   * @param {string} serverUrl URL of the Engine used to server the database.
   * @param {string} dbName Database name.
   * @param {Object=} opt_options Advanced options of this class.
   * @param {number} opt_options.startupTimeout Time limit in milliseconds used to wait the Engine
   * startup.
   * @param {number} opt_options.httpPort HTTP port to be used by the new Engine process. If not
   * informed, a random port will be used.
   * @param {Array<string>} opt_options.jazArchives JAZ files to be included into
   * Union File System (UFS).
   * @param {Array<string>} opt_options.ufsDirectories Directories to be included into
   * Union File System (UFS).
   * @param {boolean} opt_options.disableIdoSyncToDisk Disable IDO sync to disk to gain writing
   * performance on DBCache. This options increases the risk of dbCache data corruption if the
   * Engine is stopped forcefully.
   */
  constructor (installPath, serverUrl, dbName, opt_options) {
    const options = opt_options || {};
    this.serverUrl = serverUrl;
    this.dbName = dbName;
    this.installPath = installPath;
    this.enginePath = path.join(this.installPath, ENGINE_APP);
    this.process_ = null;
    this.lastStatusCode_ = 0;
    this.monitorEnabled_ = false;
    this.monitorTimeoutId_ = null;
    this.httpPort = options.httpPort || 0;
    this.useRandomHttpPort_ = !this.httpPort;
    this.jazArchives_ = options.jazArchives || [];
    this.ufsDirectories_ = options.ufsDirectories || [];
    this.disableIdoSyncToDisk_ = options.disableIdoSyncToDisk || false;
    this.startupTimeout = options.startupTimeout || ENGINE_STARTUP_TIMEOUT;
    this.startedSuccessfully = false;
  }

  /**
   * Returns true if the Engine is installed in the informed install path.
   * @return {boolean}
   */
  get installed() {
    return fs.existsSync(this.enginePath);
  }

  /**
   * Returns true if the Engine was started and if it is capable to respond HTTP requests.
   * @return {boolean}
   */
  get running() {
    return this.process_ !== null && this.lastStatusCode_ === 200;
  }

  /**
   * URL that can be used to do HTTP requests to this Engine.
   * @return {string}
   */
  get url() {
    return 'http://localhost:' + this.httpPort;
  }

  /**
   * Starts the Engine status monitor, used to discover if the Engine is running.
   * @private
   */
  startStatusMonitor_() {
    if (this.monitorTimeoutId_) {
      clearTimeout(this.monitorTimeoutId_);
      this.monitorTimeoutId_ = null;
    }
    this.monitorEnabled_ = true;
    request.get(url.format({
      protocol: 'http',
      hostname: 'localhost',
      port: this.httpPort,
      pathname: ENGINE_STATUS_URL
    }), (error, response) => {
      if (response && response.statusCode) {
        this.lastStatusCode_ = response.statusCode;
        if (this.lastStatusCode_ === 500) {
          console.info(`Engine response (${response.statusCode}):\n${response.body}`);
        }
      } else {
        this.lastStatusCode_ = null;
      }
      if (this.monitorEnabled_) {
        this.monitorTimeoutId_ = setTimeout(this.startStatusMonitor_.bind(this),
            STATUS_MONITOR_INTERVAL);
      }
    });
  }

  /**
   * Stops the Engine status monitor, started previously by {@link #startStatusMonitor_}.
   * @private
   */
  stopStatusMonitor_() {
    this.monitorEnabled_ = false;
    if (this.monitorTimeoutId_) {
      clearTimeout(this.monitorTimeoutId_);
      this.monitorTimeoutId_ = null;
    }
  }

  /**
   * Downloads the Engine and install it into the informed install path.
   * @return {Promise}
   */
  install() {
    return new Promise((resolve, reject) => {
      if (this.installed) {
        resolve();
      } else {
        mkdirp(this.installPath, (err) => {
          if (err && err.code !== 'EEXIST') {
            reject(err);
          } else {
            request.get(url.resolve(this.serverUrl, ENGINE_DOWNLOAD_URL))
            .pipe(fs.createWriteStream(this.enginePath))
            .on('close', () => {
              resolve();
            })
            .on('error', (err) => {
              reject(err);
            });
          }
        });
      }
    });
  }

  /**
   * Remove all files of the informed install path.
   * @return {Promise}
   */
  uninstall() {
    return del(path.join(this.installPath, '**'));
  }

  /**
   * Starts the Engine and wait for it to be able to handle HTTP requests. It will wait the full
   * load and synchronization of the Engine DBCache.
   * @return {Promise}
   */
  start() {
    return this.install().then(() => {
      if (this.useRandomHttpPort_) {
        return getPort();
      } else {
        return Promise.resolve(this.httpPort);
      }
    }).then(port => {
      this.httpPort = port;
      return new Promise((resolve, reject) => {
        console.info('Starting Engine process at port', this.httpPort);
        const args = [this.serverUrl, this.dbName,
          '--disableAutoRestart',
          '--disableDBCacheBackup',
          '--disableGui',
          '--httpPort=' + this.httpPort];
        this.jazArchives_.forEach((path) => {
          args.push('--jazArchive=\'' + path + '\'');
        });
        this.ufsDirectories_.forEach((path) => {
          args.push('--ufsDirectory=\'' + path + '\'');
        });
        if (this.disableIdoSyncToDisk_) {
          args.push('--disableIdoSyncToDisk');
        }
        this.process_ = spawn(ENGINE_APP, args, {
          cwd: this.installPath
        });

        this.process_.on('error', (err) => {
          console.error('Error starting Engine process:', err);
        });

        this.process_.stdout.on('data', (data) => {
          console.info(`Engine stdout: ${data}`);
        });

        this.process_.stderr.on('data', (data) => {
          console.error(`Engine stderr: ${data}`);
        });

        let intervalId = null;

        this.process_.on('close', (code) => {
          this.stopStatusMonitor_();
          if (intervalId) {
            clearInterval(intervalId);
            intervalId = null;
          }
          console.info(`Engine process exited with code ${code || 0}`);
          this.process_ = null;
          this.lastStatusCode_ = 0;
          this.httpPort = 0;
          if (code === EXIT_CODE_RESTART_REQUIRED) {
            console.info('Engine restart was required to finish the upgrade');
            sleep(UPDATE_INTERVAL);
            this.start().then(resolve, reject);
          } else {
            if (!this.startedSuccessfully) {
              const msg = `Engine startup failed with error code ${code || 0}`;
              console.error(msg);
              reject(new Error(msg));
            }
          }
          this.startedSuccessfully = false;
        });

        this.startStatusMonitor_();
        intervalId = setInterval(() => {
          if (this.process_ && this.monitorEnabled_) {
            if (this.lastStatusCode_) {
              if (this.lastStatusCode_ === 200) {
                this.startedSuccessfully = true;
                clearInterval(intervalId);
                intervalId = null;
                resolve();
              } else if (this.lastStatusCode_ === 500) {
                clearInterval(intervalId);
                intervalId = null;
                const msg = 'Unexpected Engine status code: ' + this.lastStatusCode_;
                console.error(msg);
                console.info('Kill signal sent to Engine process');
                this.process_.kill();
                reject(new Error(msg));
              }
            }
          } else {
            if (intervalId) {
              clearInterval(intervalId);
              intervalId = null;
            }
          }
        }, STATUS_MONITOR_INTERVAL);
      });
    });
  }

  /**
   * Stops the Engine process.
   * @return {Promise}
   */
  stop() {
    return new Promise(resolve => {
      if (this.running) {
        this.process_.on('close', () => {
          console.info('Engine stopped');
          resolve();
        });
        console.info('Kill signal sent to Engine process');
        this.process_.kill();
        this.process_ = null;
      } else {
        resolve();
      }
    });
  }
}

module.exports = EngineWrapper;
